package refereeview;

import java.sql.*;

public class Database {
    private Connection fConnection = null;
    private int fEventId = -1;   // Read-only; provided by the database

    // Default constructor -- useless (should probably throw an exception)
    private Database() { }

    // Connecting constructor -- use this one
    // port will usually be 3306 for normal MySQL servers
    public Database(String host, short port, String user, String password,
                    String dbname) throws Exception {
        RegisterMySQLDriver();
        fConnection = DriverManager.getConnection(
            "jdbc:mysql://" + host + "/" + dbname + "?user=" + user +
            "&password=" + password);
    }

    // Creates and registers the JDBC Driver.
    private static boolean isDriverRegistered = false;
    private static void RegisterMySQLDriver() throws Exception {
        if (!isDriverRegistered) {
            // Only called once per execution
            Class.forName("com.mysql.jdbc.Driver");
            isDriverRegistered = true;
        }
    }

    // Accessors and Mutators
    public int getEventId() {
        return fEventId;
    }

    // Inserts a new competitor into the database (e.g., registration)
    public void addCompetitor(Competitor comp) throws Exception {
        if (fEventId == -1)
            throw new Exception("No Event specified");

        Statement statement = fConnection.createStatement();
        statement.executeUpdate("INSERT INTO `Competitors` SET " +
            "`EventID`='" + fEventId + "', " +
            "`FName`='" + escapeString(comp.getFirstName()) + "', " +
            "`LName`='" + escapeString(comp.getLastName()) + "', " +
            "`BeltRank`='" + escapeString(comp.getBeltRank()) + "', " +
            "`Age`='" + escapeString(comp.getAge()) + "', " +
            "`Matches`='0', `Losses`='0', `Passes`='0', `Flags`='0'",
            Statement.RETURN_GENERATED_KEYS);

        // Update the Competitor object with the database ID
        ResultSet keys = statement.getGeneratedKeys();
        if (keys.first())
            comp.setId(keys.getInt(1));
        else
            throw new Exception("Error adding competitor");

        keys.close();
        statement.close();
    }

    // Updates the information for an existing competitor
    public void updateCompetitor(Competitor comp) throws Exception {
        if (fEventId == -1)
            throw new Exception("No Event specified");

        Statement statement = fConnection.createStatement();
        statement.execute("UPDATE `Competitors` SET " +
            "`FName`='" + escapeString(comp.getFirstName()) + "', " +
            "`LName`='" + escapeString(comp.getLastName()) + "', " +
            "`BeltRank`='" + escapeString(comp.getBeltRank()) + "', " +
            "`Age`='" + escapeString(comp.getAge()) + "' " +
            "WHERE `id`='" + comp.getId() + "'");
        statement.close();
    }

    // Remove a competitor from the database
    public void delCompetitor(Competitor comp) throws Exception {
        if (fEventId == -1)
            throw new Exception("No Event specified");

        Statement statement = fConnection.createStatement();
        statement.execute("DELETE FROM `Competitors` " +
            "WHERE `id`='" + comp.getId() + "'");
        statement.close();
    }

    // Find a competitor based on search criteria.
    public Competitor findCompetitor(String firstName, String lastName) throws Exception {
        if (fEventId == -1)
            throw new Exception("No Event specified");

        Statement statement = fConnection.createStatement();
        ResultSet result = statement.executeQuery("SELECT * FROM `Competitors` " +
            "WHERE `EventID`='" + fEventId + "' " +
            "AND `FName` LIKE '" + escapeString(firstName) + "' " +
            "AND `LName` LIKE '" + escapeString(lastName) + "'");

        Competitor found = null;
        if (result.first())
            // A competitor was found.  Yay!
            found = getCompetitor(result);

        result.close();
        statement.close();
        return found;
    }
    
    // Get a competitor object from a database result
    private Competitor getCompetitor(ResultSet res) throws SQLException {
        Competitor comp = new Competitor(res.getInt("id"));
        comp.setName(stripSlashes(res.getString("FName")), stripSlashes(res.getString("LName")));
        comp.setBeltRank(stripSlashes(res.getString("BeltRank")));
        comp.setAge(stripSlashes(res.getString("Age")));
        comp.setMatches(res.getInt("Matches"));
        comp.setLosses(res.getInt("Losses"));
        comp.setPasses(res.getInt("Passes"));
        comp.setFlags(res.getInt("Flags"));
        return comp;
    }
    
    // Create an event.  This database will be linked to the event until
    // the object is destroyed, or until a new event is created
    public void createEvent(String description) throws Exception {
        Statement statement = fConnection.createStatement();
        statement.executeUpdate("INSERT INTO `Events` SET " +
            "`Description`='" + escapeString(description) + "'",
            Statement.RETURN_GENERATED_KEYS);

        // Store the Event ID
        ResultSet keys = statement.getGeneratedKeys();
        if (keys.first())
            fEventId = keys.getInt(1);
        else
            throw new Exception("Error creating event");

        keys.close();
        statement.close();
    }

    // Join an existing event.  The event description will be looked up
    // in the database, and the Event ID will be set accordingly.  If the
    // event does not already exist, createEvent will be called to add it.
    public void joinEvent(String desc) throws Exception {
        Statement statement = fConnection.createStatement();
        ResultSet result = statement.executeQuery("SELECT * FROM `Events` " +
            "WHERE `Description` LIKE '" + escapeString(desc) + "'");

        fEventId = -1;
        if (result.first())
            fEventId = result.getInt("id");
        result.close();
        statement.close();
        
        // We didn't find the event, so let's add it
        if (fEventId == -1)
            createEvent(desc);
    }

    // Create a match
    // Selects 2 competitors and returns a Match object, which can be used in
    // the other match functions below
    public Match createMatch() throws Exception {
        if (fEventId == -1)
            throw new Exception("No Event specified");

        // First see if we have two competitors left:
        Statement statement = fConnection.createStatement();
        Competitor comp1, comp2;
        ResultSet result = statement.executeQuery(
            "SELECT COUNT(*) FROM `Competitors` WHERE `Losses`='0' AND `Flags`='0'");
        if (!result.first() || result.getInt(1) < 2)
            return null;
        
        // Good, we do, so see if we can create a match with them...
        result = statement.executeQuery("SELECT * FROM `Competitors` " +
            "WHERE `EventID`='" + fEventId + "' AND `Losses`='0' AND `Flags`='0' " +
            "ORDER BY `Matches` ASC, `Passes` ASC");
        if (!result.first())
            return null;
        comp1 = getCompetitor(result);

        // We have the first, so lets look for the next
        result = statement.executeQuery("SELECT * FROM `Competitors` " +
            "WHERE `EventID`='" + fEventId + "' AND `Losses`='0' AND `Flags`='0' " +
            "AND `BeltRank`='" + comp1.getBeltRank() + "' " +
            "AND `Age`='" + comp1.getAge() + "' " +
            //"AND `Matches`='" + comp1.getMatches() + "' " +
            "AND NOT(`id`='" + comp1.getId() + "') " +
            "ORDER BY `Matches` ASC, `Passes` ASC");
        
        if (result.first()) {
            // Got a second competitor -- do a normal match
            comp2 = getCompetitor(result);
            comp1.doMatch();
            comp2.doMatch();
            statement.execute("UPDATE `Competitors` " +
                "SET `Matches`='" + comp1.getMatches() + "' " +
                "WHERE `id`='" + comp1.getId() + "'");
            statement.execute("UPDATE `Competitors` " +
                "SET `Matches`='" + comp2.getMatches() + "' " +
                "WHERE `id`='" + comp2.getId() + "'");
        } else {
            // No second competitor -- competitor 1 gets a pass, and we start over again
            comp1.doMatch();
            comp1.pass();
            statement.execute("UPDATE `Competitors` " +
                "SET `Matches`='" + comp1.getMatches() + "', " +
                "`Passes`='" + comp1.getPasses() + "' " +
                "WHERE `id`='" + comp1.getId() + "'");
            statement.execute("INSERT INTO `Matches` SET `EventID`='" + fEventId + "', " +
                "`Competitor1`='" + comp1.getId() + "', " +
                "`Competitor2`='0', `Winner`='1', `Time`=CURRENT_TIMESTAMP()");
            result.close();
            statement.close();
            return createMatch();
        }

        // Create the match itself
        statement.executeUpdate("INSERT INTO `Matches` SET `EventID`='" + fEventId + "', " +
            "`Competitor1`='" + comp1.getId() + "', " +
            "`Competitor2`='" + comp2.getId() + "', " + 
            "`Winner`='0', `Time`=CURRENT_TIMESTAMP()",
            Statement.RETURN_GENERATED_KEYS);
        ResultSet keys = statement.getGeneratedKeys();
        int id = -1;
        if (keys.first())
            id = keys.getInt(1);
        
        result.close();
        keys.close();
        statement.close();
        return new Match(id, comp1, comp2);
    }

    // Notifies the database of the winner of a match.  If winner is either 1
    // or 2, it indicates which competitor won from the match.  If winner is
    // 0, it specifies that the match had only one competitor, and that
    // competitor should get a pass.
    public void setWinner(Match match, int winner) throws Exception {
        if (fEventId == -1)
            throw new Exception("No Event specified");

        Statement statement = fConnection.createStatement();
        
        if (winner == 1) {
            match.getCompetitor2().youLose();
            statement.execute("UPDATE `Matches` " +
                "SET `Winner`='" + match.getCompetitor1().getId() + "' "+
                "WHERE `id`='" + match.getMatchId() + "'");
            statement.execute("UPDATE `Competitors` " +
                "SET `Losses`='" + match.getCompetitor2().getLosses() + "', " +
                "`Flags`='" + match.getCompetitor2().getFlags() + "' " +
                "WHERE `id`='" + match.getCompetitor2().getId() + "'");
            statement.execute("UPDATE `Competitors` " +
                "SET `Flags`='" + match.getCompetitor1().getFlags() + "' " +
                "WHERE `id`='" + match.getCompetitor1().getId() + "'");
        }
        else if (winner == 2) {
            match.getCompetitor1().youLose();
            statement.execute("UPDATE `Matches` " +
                "SET `Winner`='" + match.getCompetitor2().getId() + "' "+
                "WHERE `id`='" + match.getMatchId() + "'");
            statement.execute("UPDATE `Competitors` " +
                "SET `Losses`='" + match.getCompetitor1().getLosses() + "', " +
                "`Flags`='" + match.getCompetitor1().getFlags() + "' " +
                "WHERE `id`='" + match.getCompetitor1().getId() + "'");
            statement.execute("UPDATE `Competitors` " +
                "SET `Flags`='" + match.getCompetitor2().getFlags() + "' " +
                "WHERE `id`='" + match.getCompetitor2().getId() + "'");
        }
        else {
            throw new Exception("Invalid winner specified!");
        }
    }
    
    // Calculate the depth of a match tree
    private int scanMatchDepth(Statement statement, int winnerId, String time) throws SQLException {
        ResultSet result = statement.executeQuery("SELECT * FROM `Matches` " +
            "WHERE `EventID`='" + fEventId + "' AND `Winner`='" + winnerId + "' " +
            "AND `Time`<'" + time + "'");
        if (!result.first()) {
            result.close();
            return 1;
        }
        int leftId = result.getInt("Competitor1");
        int rightId = result.getInt("Competitor2");
        String newTime = result.getString("Time");
        
        int left = scanMatchDepth(statement, leftId, newTime);
        int right = scanMatchDepth(statement, rightId, newTime);
        result.close();
        return (left > right) ? left : right;
    }
    
    // Return the actual matches in a "tree" format
    private void getMatches(Statement statement, int winnerId, String time,
                            int x, int y, String[][] output) throws SQLException {
        ResultSet result = statement.executeQuery("SELECT * FROM `Matches` " +
            "WHERE `EventID`='" + fEventId + "' AND `Winner`='" + winnerId + "' " +
            "AND `Time`<'" + time + "'");
        if (!result.first()) {
            result.close();
            return;
        }
        
        int idLeft = result.getInt("Competitor1");
        int idRight = result.getInt("Competitor2");
        String newTime = result.getString("Time");
        
        if (idLeft > 0) {
            result = statement.executeQuery("SELECT * FROM `Competitors` WHERE `id`='" + idLeft + "'");
            if (result.first()) {
                Competitor comp = getCompetitor(result);
                output[x-1][y*2] = comp.getFullName();
            }
            getMatches(statement, idLeft, newTime, x-1, y*2, output);
        } else {
            output[x-1][y*2] = null;
        }

        if (idRight > 0) {
            result = statement.executeQuery("SELECT * FROM `Competitors` WHERE `id`='" + idRight + "'");
            if (result.first()) {
                Competitor comp = getCompetitor(result);
                output[x-1][(y*2)+1] = comp.getFullName();
            }
            getMatches(statement, idRight, newTime, x-1, (y*2)+1, output);
        } else {
            output[x-1][(y*2)+1] = null;
        }
        result.close();
    }
    
    public String[][] getBracket() throws SQLException {
        Statement statement = fConnection.createStatement();
        ResultSet result = statement.executeQuery("SELECT * FROM `Matches` " +
            "WHERE `EventID`='" + fEventId + "' AND NOT(`Winner`='0') " +
            "ORDER BY `Time` DESC");
        if (!result.first())
            return null;
        
        String[][] bracket = new String[12][12];
        int winId = result.getInt("Winner");
        String timeLast = "9999-12-31 23:59:59";
        int width = scanMatchDepth(statement, winId, timeLast) + 1;
        
        result = statement.executeQuery("SELECT * FROM `Competitors` WHERE `id`='" + winId + "'");
        if (!result.first())
            return null;
        Competitor comp = getCompetitor(result);
        bracket[width][0] = comp.getFullName();
        getMatches(statement, winId, timeLast, width, 0, bracket);
        
        result.close();
        statement.close();
        return bracket;
    }

    // Cleans malicious or problematic characters from an input string
    // by escaping them with backslashes
    private static String escapeString(String str) {
        // Regex:  Replace ', ", \ with \', \", \\
        return str.replaceAll("(['\"\\\\])", "\\\\$1");
    }
    
    // Reverse the work of the escapeString function
    private static String stripSlashes(String str) {
        return str.replaceAll("\\\\", "");
    }
}
